from dataclasses import dataclass
from typing import Sequence

from ..operation import Operation
from ..specification import ClientSpecification
from ..typing import Dependency
from .api import (
    CommonDecorator,
    ConsumerDecorator,
    SupportsConsumeClientSpecification,
    SupportsConsumeOperation,
)

__all__: Sequence[str] = ("response",)


@dataclass
class ResponseConsumer(SupportsConsumeOperation, SupportsConsumeClientSpecification):
    dependency: Dependency

    def consume_operation(self, operation: Operation, /) -> None:
        operation.response = self.dependency

    def consume_client_spec(self, client_specification: ClientSpecification, /) -> None:
        client_specification.default_response = self.dependency


def response(dependency: Dependency, /) -> CommonDecorator:
    return ConsumerDecorator(ResponseConsumer(dependency))



from dataclasses import dataclass, field
from typing import Optional, Sequence

import mediate
import mediatype
from mediatype import MediaType

from .auths import Auth
from .enums import HTTPHeader
from .errors import (
    ExpectedContentTypeError,
    ExpectedHeaderError,
    ExpectedStatusCodeError,
)
from .models import Request, Response
from .typing import CallNext, MiddlewareCallable

__all__: Sequence[str] = (
    "Middleware",
    "AuthMiddleware",
    "ExpectedStatusCodeMiddleware",
    "ExpectedHeaderMiddleware",
    "ExpectedContentTypeMiddleware",
    "raise_for_status",
)


class Middleware(mediate.Middleware[Request, Response]):
    pass


@dataclass
class AuthMiddleware(MiddlewareCallable):
    auth: Auth

    def __call__(self, call_next: CallNext, request: Request, /) -> Response:
        return call_next(self.auth.auth(request))


@dataclass(init=False)
class ExpectedStatusCodeMiddleware:
    codes: Sequence[int] = field(default_factory=list)

    def __init__(self, *codes: int) -> None:
        self.codes = codes

    def __call__(self, call_next: CallNext, request: Request, /) -> Response:
        response: Response = call_next(request)

        if response.status_code not in self.codes:
            raise ExpectedStatusCodeError(
                f"Response contained an unexpected status code: {response.status_code}"
            )

        return response


@dataclass
class ExpectedHeaderMiddleware:
    name: str
    value: Optional[str] = None

    def __call__(self, call_next: CallNext, request: Request, /) -> Response:
        response: Response = call_next(request)

        if self.name not in response.headers:
            raise ExpectedHeaderError(name=self.name)

        header_value: str = response.headers[self.name]

        if self.value is not None and header_value != self.value:
            raise ExpectedHeaderError(
                name=self.name, value=header_value, expected_value=self.value
            )

        return response


@dataclass
class ExpectedContentTypeMiddleware:
    content_type: MediaType

    suffix: bool
    parameters: bool

    def __init__(
        self,
        content_type: str,
        /,
        *,
        suffix: Optional[bool] = None,
        parameters: Optional[bool] = None,
    ) -> None:
        self.content_type = mediatype.parse(content_type)

        if suffix is None:
            suffix = self.content_type.suffix is not None
        if parameters is None:
            parameters = self.content_type.parameters is not None

        self.suffix = suffix
        self.parameters = parameters

    def __call__(self, call_next: CallNext, request: Request, /) -> Response:
        response: Response = call_next(request)

        raw_content_type: str = response.headers.get(HTTPHeader.CONTENT_TYPE)

        if raw_content_type is None:
            raise ExpectedHeaderError(HTTPHeader.CONTENT_TYPE)

        content_type: MediaType = mediatype.parse(raw_content_type)

        expected_content_type: str = self._media_type_to_string(self.content_type)
        actual_content_type: str = self._media_type_to_string(content_type)

        if expected_content_type != actual_content_type:
            raise ExpectedContentTypeError(
                expected=expected_content_type, actual=actual_content_type
            )

        return response

    def _media_type_to_string(self, media_type: MediaType, /) -> str:
        return media_type.string(suffix=self.suffix, parameters=self.parameters)


def raise_for_status(call_next: CallNext, request: Request, /) -> Response:
    response: Response = call_next(request)

    response.raise_for_status()

    return response



from typing import Sequence

from ..auths import Auth, BasicAuth
from ..middlewares import AuthMiddleware
from .api import CommonDecorator
from .middleware import middleware

__all__: Sequence[str] = (
    "auth",
    "basic_auth",
)


def auth(auth: Auth, /) -> CommonDecorator:
    return middleware(AuthMiddleware(auth))


def basic_auth(username: str, password: str) -> CommonDecorator:
    return auth(BasicAuth(username, password))



from typing import Mapping, Sequence

from ..consumers import (
    ContentConsumer,
    DataConsumer,
    FilesConsumer,
    JsonConsumer,
    MountConsumer,
    PathConsumer,
    PathParamsConsumer,
)
from ..converters import convert_path_param, convert_path_params
from ..types import (
    JsonTypes,
    PathParamsTypes,
    PathTypes,
    RequestContent,
    RequestData,
    RequestFiles,
)
from .api import OperationConsumerDecorator, OperationDecorator

__all__: Sequence[str] = (
    "content",
    "data",
    "files",
    "json",
    "mount",
    "path",
    "path_params",
)


def content(content: RequestContent, /) -> OperationDecorator:
    return OperationConsumerDecorator(ContentConsumer(content))


def data(data: RequestData, /) -> OperationDecorator:
    return OperationConsumerDecorator(DataConsumer(data))


def files(files: RequestFiles, /) -> OperationDecorator:
    return OperationConsumerDecorator(FilesConsumer(files))


def json(json: JsonTypes, /) -> OperationDecorator:
    return OperationConsumerDecorator(JsonConsumer(json))


def mount(path: str, /) -> OperationDecorator:
    return OperationConsumerDecorator(MountConsumer(path))


def path(
    key: str,
    value: PathTypes,
    *,
    delimiter: str = "/",
) -> OperationDecorator:
    converted_value: str = convert_path_param(value, delimiter=delimiter)

    return OperationConsumerDecorator(PathConsumer(key, converted_value))


def path_params(
    path_params: PathParamsTypes,
    /,
    *,
    delimiter: str = "/",
) -> OperationDecorator:
    converted_path_params: Mapping[str, str] = convert_path_params(
        path_params, delimiter=delimiter
    )

    return OperationConsumerDecorator(PathParamsConsumer(converted_path_params))



from http.cookiejar import CookieJar
from typing import Mapping, MutableMapping, MutableSequence, Sequence

from httpx import Cookies, Headers, QueryParams, Timeout
from httpx._utils import primitive_value_to_str

from .errors import ConversionError
from .types import (
    CookiesTypes,
    CookieTypes,
    HeadersTypes,
    HeaderTypes,
    PathParamsTypes,
    PathTypes,
    Primitive,
    QueryParamsTypes,
    QueryTypes,
    TimeoutTypes,
)
from .utils import is_primitive

__all__: Sequence[str] = (
    "convert_query_param",
    "convert_header",
    "convert_cookie",
    "convert_path_param",
    "convert_query_params",
    "convert_headers",
    "convert_cookies",
    "convert_path_params",
    "convert_timeout",
)


def convert_query_param(value: QueryTypes, /) -> Sequence[str]:
    if isinstance(value, (str, int, float, bool)) or value is None:
        return [primitive_value_to_str(value)]

    if isinstance(value, Sequence):
        return [primitive_value_to_str(item) for item in value]

    raise ConversionError("query param", value)


def convert_header(value: HeaderTypes, /) -> Sequence[str]:
    if isinstance(value, (str, int, float, bool)) or value is None:
        return [primitive_value_to_str(value)]

    if isinstance(value, Sequence):
        return [primitive_value_to_str(item) for item in value]

    raise ConversionError("header", value)


def convert_cookie(value: CookieTypes, /) -> str:
    if is_primitive(value):
        return primitive_value_to_str(value)

    return str(value)


def convert_path_param(value: PathTypes, /, *, delimiter: str = "/") -> str:
    if isinstance(value, (str, int, float, bool)) or value is None:
        return primitive_value_to_str(value)

    if isinstance(value, Sequence):
        segments: MutableSequence = []

        segment: Primitive
        for segment in value:
            converted_segment: str = primitive_value_to_str(segment)

            if converted_segment:
                segments.append(converted_segment)

        return delimiter.join(segments)

    raise ConversionError("path param", value)


def convert_query_params(value: QueryParamsTypes, /) -> QueryParams:
    if isinstance(value, QueryParams):
        return value

    if isinstance(value, (Mapping, str, bytes)):
        return QueryParams(value)

    if isinstance(value, Sequence):
        if all(isinstance(v, tuple) for v in value):
            return QueryParams([*value])
        elif all(isinstance(v, str) for v in value):
            return QueryParams([(str(v), None) for v in value])

    raise ConversionError("query params", value)


def convert_headers(value: HeadersTypes, /) -> Headers:
    if isinstance(value, Headers):
        return value

    if isinstance(value, (Mapping, Sequence)):
        return Headers(dict(value))

    raise ConversionError("headers", value)


def convert_cookies(value: CookiesTypes, /) -> Cookies:
    if isinstance(value, Cookies):
        return value

    if isinstance(value, CookieJar):
        return Cookies(value)

    if isinstance(value, (Mapping, Sequence)):
        return Cookies(dict(value))

    raise ConversionError("cookies", value)


def convert_path_params(
    path_params: PathParamsTypes, /, *, delimiter: str = "/"
) -> MutableMapping[str, str]:
    if isinstance(path_params, Mapping):
        return {
            key: convert_path_param(value, delimiter=delimiter)
            for key, value in path_params.items()
        }

    return {
        key: convert_path_param(value, delimiter=delimiter)
        for key, value in path_params
    }


def convert_timeout(value: TimeoutTypes, /) -> Timeout:
    return Timeout(value)



from typing import Any, Callable, Optional, Sequence, TypeVar

from pydantic.fields import Undefined

from .dependence import DependencyParameter
from .params import (
    AllRequestStateParameter,
    AllResponseStateParameter,
    AllStateParameter,
    BodyParameter,
    CookieParameter,
    CookiesParameter,
    HeaderParameter,
    HeadersParameter,
    Parameter,
    PathParameter,
    PathParamsParameter,
    QueryParameter,
    QueryParamsParameter,
    ReasonParameter,
    RequestParameter,
    ResponseParameter,
    StateParameter,
    StatusCodeParameter,
    URLParameter,
)
from .typing import Supplier

__all__: Sequence[str] = (
    "Query",
    "Header",
    "Cookie",
    "Path",
    "QueryParams",
    "Headers",
    "Cookies",
    "PathParams",
    "Body",
    "Depends",
    "URL",
    "Reason",
    "Request",
    "Response",
    "StatusCode",
    "State",
    "AllRequestState",
    "AllResponseState",
    "AllState",
)

P = TypeVar("P", bound=Parameter)


def _validate(parameter: P, /) -> P:
    parameter._validate()

    return parameter


def Query(
    name: Optional[str] = None,
    *,
    default: Any = Undefined,
    default_factory: Optional[Supplier[Any]] = None,
    const: Optional[bool] = None,
    gt: Optional[float] = None,
    ge: Optional[float] = None,
    lt: Optional[float] = None,
    le: Optional[float] = None,
    multiple_of: Optional[float] = None,
    allow_inf_nan: Optional[bool] = None,
    max_digits: Optional[int] = None,
    decimal_places: Optional[int] = None,
    min_length: Optional[int] = None,
    max_length: Optional[int] = None,
    regex: Optional[str] = None,
) -> QueryParameter:
    return _validate(
        QueryParameter(
            default=default,
            default_factory=default_factory,
            alias=name,
            const=const,
            gt=gt,
            ge=ge,
            lt=lt,
            le=le,
            multiple_of=multiple_of,
            allow_inf_nan=allow_inf_nan,
            max_digits=max_digits,
            decimal_places=decimal_places,
            min_length=min_length,
            max_length=max_length,
            regex=regex,
        )
    )


def Header(
    name: Optional[str] = None,
    *,
    default: Any = Undefined,
    default_factory: Optional[Supplier[Any]] = None,
    const: Optional[bool] = None,
    gt: Optional[float] = None,
    ge: Optional[float] = None,
    lt: Optional[float] = None,
    le: Optional[float] = None,
    multiple_of: Optional[float] = None,
    allow_inf_nan: Optional[bool] = None,
    max_digits: Optional[int] = None,
    decimal_places: Optional[int] = None,
    min_length: Optional[int] = None,
    max_length: Optional[int] = None,
    regex: Optional[str] = None,
    convert_underscores: bool = True,
) -> HeaderParameter:
    return _validate(
        HeaderParameter(
            default=default,
            default_factory=default_factory,
            alias=name,
            const=const,
            gt=gt,
            ge=ge,
            lt=lt,
            le=le,
            multiple_of=multiple_of,
            allow_inf_nan=allow_inf_nan,
            max_digits=max_digits,
            decimal_places=decimal_places,
            min_length=min_length,
            max_length=max_length,
            regex=regex,
            convert_underscores=convert_underscores,
        )
    )


def Cookie(
    name: Optional[str] = None,
    *,
    default: Any = Undefined,
    default_factory: Optional[Supplier[Any]] = None,
    const: Optional[bool] = None,
    gt: Optional[float] = None,
    ge: Optional[float] = None,
    lt: Optional[float] = None,
    le: Optional[float] = None,
    multiple_of: Optional[float] = None,
    allow_inf_nan: Optional[bool] = None,
    max_digits: Optional[int] = None,
    decimal_places: Optional[int] = None,
    min_length: Optional[int] = None,
    max_length: Optional[int] = None,
    regex: Optional[str] = None,
) -> CookieParameter:
    return _validate(
        CookieParameter(
            default=default,
            default_factory=default_factory,
            alias=name,
            const=const,
            gt=gt,
            ge=ge,
            lt=lt,
            le=le,
            multiple_of=multiple_of,
            allow_inf_nan=allow_inf_nan,
            max_digits=max_digits,
            decimal_places=decimal_places,
            min_length=min_length,
            max_length=max_length,
            regex=regex,
        )
    )


def Path(
    name: Optional[str] = None,
    *,
    default: Any = Undefined,
    default_factory: Optional[Supplier[Any]] = None,
    const: Optional[bool] = None,
    gt: Optional[float] = None,
    ge: Optional[float] = None,
    lt: Optional[float] = None,
    le: Optional[float] = None,
    multiple_of: Optional[float] = None,
    allow_inf_nan: Optional[bool] = None,
    max_digits: Optional[int] = None,
    decimal_places: Optional[int] = None,
    min_length: Optional[int] = None,
    max_length: Optional[int] = None,
    regex: Optional[str] = None,
    delimiter: str = "/",
) -> PathParameter:
    return _validate(
        PathParameter(
            default=default,
            default_factory=default_factory,
            alias=name,
            const=const,
            gt=gt,
            ge=ge,
            lt=lt,
            le=le,
            multiple_of=multiple_of,
            allow_inf_nan=allow_inf_nan,
            max_digits=max_digits,
            decimal_places=decimal_places,
            min_length=min_length,
            max_length=max_length,
            regex=regex,
            delimiter=delimiter,
        )
    )


def QueryParams(
    *,
    default: Any = Undefined,
    default_factory: Optional[Supplier[Any]] = None,
    min_items: Optional[int] = None,
    max_items: Optional[int] = None,
    unique_items: Optional[bool] = None,
) -> QueryParamsParameter:
    return _validate(
        QueryParamsParameter(
            default=default,
            default_factory=default_factory,
            min_items=min_items,
            max_items=max_items,
            unique_items=unique_items,
        )
    )


def Headers(
    *,
    default: Any = Undefined,
    default_factory: Optional[Supplier[Any]] = None,
    min_items: Optional[int] = None,
    max_items: Optional[int] = None,
    unique_items: Optional[bool] = None,
) -> HeadersParameter:
    return _validate(
        HeadersParameter(
            default=default,
            default_factory=default_factory,
            min_items=min_items,
            max_items=max_items,
            unique_items=unique_items,
        )
    )


def Cookies(
    *,
    default: Any = Undefined,
    default_factory: Optional[Supplier[Any]] = None,
    min_items: Optional[int] = None,
    max_items: Optional[int] = None,
    unique_items: Optional[bool] = None,
) -> CookiesParameter:
    return _validate(
        CookiesParameter(
            default=default,
            default_factory=default_factory,
            min_items=min_items,
            max_items=max_items,
            unique_items=unique_items,
        )
    )


def PathParams(
    *,
    default: Any = Undefined,
    default_factory: Optional[Supplier[Any]] = None,
    min_items: Optional[int] = None,
    max_items: Optional[int] = None,
    unique_items: Optional[bool] = None,
    delimiter: str = "/",
) -> PathParamsParameter:
    return _validate(
        PathParamsParameter(
            default=default,
            default_factory=default_factory,
            min_items=min_items,
            max_items=max_items,
            unique_items=unique_items,
            delimiter=delimiter,
        )
    )


def Body(
    *,
    default: Any = Undefined,
    default_factory: Optional[Supplier[Any]] = None,
    alias: Optional[str] = None,
    embed: bool = False,
) -> BodyParameter:
    return _validate(
        BodyParameter(
            alias=alias,
            default=default,
            default_factory=default_factory,
            embed=embed,
        )
    )


def Depends(
    dependency: Optional[Callable] = None,
    /,
    *,
    use_cache: bool = True,
) -> DependencyParameter:
    return _validate(
        DependencyParameter(
            dependency=dependency,
            use_cache=use_cache,
        )
    )


def URL() -> URLParameter:
    return _validate(URLParameter())


def Request() -> RequestParameter:
    return _validate(RequestParameter())


def Response() -> ResponseParameter:
    return _validate(ResponseParameter())


def StatusCode() -> StatusCodeParameter:
    return _validate(StatusCodeParameter())


def Reason() -> ReasonParameter:
    return _validate(ReasonParameter())


def State(
    name: Optional[str] = None,
    *,
    default: Any = Undefined,
    default_factory: Optional[Supplier[Any]] = None,
) -> StateParameter:
    return _validate(
        StateParameter(
            default=default,
            default_factory=default_factory,
            alias=name,
        )
    )


def AllRequestState() -> AllRequestStateParameter:
    return _validate(AllRequestStateParameter())


def AllResponseState() -> AllResponseStateParameter:
    return _validate(AllResponseStateParameter())


def AllState() -> AllStateParameter:
    return _validate(AllStateParameter())



from dataclasses import dataclass
from typing import Optional, Sequence

from mediate.protocols import MiddlewareCallable

from ..middlewares import (
    ExpectedContentTypeMiddleware,
    ExpectedHeaderMiddleware,
    ExpectedStatusCodeMiddleware,
)
from ..middlewares import raise_for_status as raise_for_status_middleware
from ..models import Request, Response
from ..operation import Operation
from ..specification import ClientSpecification
from .api import (
    CS,
    CommonDecorator,
    ConsumerDecorator,
    SupportsConsumeClientSpecification,
    SupportsConsumeOperation,
)

__all__: Sequence[str] = (
    "middleware",
    "expect_content_type",
    "expect_header",
    "expect_status",
    "raise_for_status",
)


@dataclass
class MiddlewareConsumer(SupportsConsumeClientSpecification, SupportsConsumeOperation):
    middleware: Sequence[MiddlewareCallable[Request, Response]]

    def consume_client_spec(self, client_specification: ClientSpecification) -> None:
        client_specification.middleware.add_all(self.middleware)

    def consume_operation(self, operation: Operation) -> None:
        operation.middleware.add_all(self.middleware)


def middleware(*middleware: MiddlewareCallable[Request, Response]) -> CommonDecorator:
    return ConsumerDecorator(MiddlewareConsumer(middleware))


def expect_content_type(content_type: str, /) -> CommonDecorator:
    return middleware(ExpectedContentTypeMiddleware(content_type))


def expect_header(name: str, value: Optional[str] = None) -> CommonDecorator:
    return middleware(ExpectedHeaderMiddleware(name, value))


def expect_status(*status_codes: int) -> CommonDecorator:
    return middleware(ExpectedStatusCodeMiddleware(*status_codes))


def raise_for_status(target: CS, /) -> CS:
    return middleware(raise_for_status_middleware)(target)



from abc import abstractmethod
from dataclasses import dataclass
from typing import (
    Any,
    Callable,
    Protocol,
    Sequence,
    Type,
    TypeVar,
    Union,
    runtime_checkable,
)

from ..errors import CompositionError
from ..models import ClientOptions, PreRequest
from ..operation import Operation, get_operation
from ..services import Service
from ..specification import ClientSpecification
from ..typing import Consumer, SupportsConsumeClient, SupportsConsumeRequest

__all__: Sequence[str] = (
    # Protocols
    "OperationDecorator",
    "ServiceDecorator",
    "CommonDecorator",
    # Wrappers
    "ConsumerDecorator",
    "OperationConsumerDecorator",
    "ServiceConsumerDecorator",
)

C = TypeVar("C", bound=Callable[..., Any])
S = TypeVar("S", bound=Type[Service])
CS = TypeVar("CS", Callable[..., Any], Type[Service])


@runtime_checkable
class SupportsConsumeClientSpecification(Protocol):
    @abstractmethod
    def consume_client_spec(self, client_specification: ClientSpecification, /) -> None:
        ...


@runtime_checkable
class SupportsConsumeOperation(Protocol):
    @abstractmethod
    def consume_operation(self, operation: Operation, /) -> None:
        ...


class ClientSpecificationConsumer(Consumer[ClientSpecification], Protocol):
    pass


class OperationConsumer(Consumer[Operation], Protocol):
    pass


class OperationDecorator(Protocol):
    @abstractmethod
    def __call__(self, target: C, /) -> C:
        raise NotImplementedError


class ServiceDecorator(Protocol):
    @abstractmethod
    def __call__(self, target: S, /) -> S:
        raise NotImplementedError


class CommonDecorator(Protocol):
    @abstractmethod
    def __call__(self, target: CS, /) -> CS:
        raise NotImplementedError


@dataclass
class OperationConsumerDecorator(OperationDecorator):
    consumer: Union[
        SupportsConsumeOperation,
        SupportsConsumeRequest,
        SupportsConsumeClient,
    ]

    def __call__(self, target: C, /) -> C:
        if not callable(target):
            raise CompositionError(f"Target of unsupported type {type(target)}")

        operation: Operation = get_operation(target)

        if isinstance(self.consumer, SupportsConsumeOperation):
            self.consumer.consume_operation(operation)
        elif isinstance(self.consumer, SupportsConsumeRequest):
            pre_request: PreRequest = operation.pre_request

            self.consumer.consume_request(pre_request)
        elif isinstance(self.consumer, SupportsConsumeClient):
            client_options: ClientOptions = operation.client_options

            self.consumer.consume_client(client_options)
        else:
            raise CompositionError(
                f"Consumer {self.consumer} does not support consuming a request or client"
            )

        return target


@dataclass
class ServiceConsumerDecorator(ServiceDecorator):
    consumer: Union[SupportsConsumeClientSpecification, SupportsConsumeClient]

    def __call__(self, target: S, /) -> S:
        if not isinstance(target, type):
            raise CompositionError(f"Target of unsupported type {type(target)}")

        if not issubclass(target, Service):
            raise CompositionError(f"Target class is not a subclass of {Service}")

        specification: ClientSpecification = target._spec
        options: ClientOptions = specification.options

        if isinstance(self.consumer, SupportsConsumeClientSpecification):
            self.consumer.consume_client_spec(specification)
        elif isinstance(self.consumer, SupportsConsumeClient):
            self.consumer.consume_client(options)
        else:
            raise CompositionError(
                f"Consumer {self.consumer} does not support consuming client specification or options"
            )

        return target


@dataclass
class ConsumerDecorator(CommonDecorator):
    consumer: Union[
        SupportsConsumeOperation,
        SupportsConsumeRequest,
        SupportsConsumeClientSpecification,
        SupportsConsumeClient,
    ]

    def __call__(self, target: CS, /) -> CS:
        if isinstance(target, type):
            if not isinstance(
                self.consumer,
                (SupportsConsumeClientSpecification, SupportsConsumeClient),
            ):
                raise CompositionError(
                    f"Consumer {type(self.consumer).__name__!r} does not support consumption"
                    f" of type {Service}"
                )

            return ServiceConsumerDecorator(self.consumer)(target)
        elif callable(target):
            if not isinstance(
                self.consumer,
                (
                    SupportsConsumeOperation,
                    SupportsConsumeRequest,
                    SupportsConsumeClient,
                ),
            ):
                raise CompositionError(
                    f"Consumer {type(self.consumer).__name__!r} does not support consuming"
                    f" target of type {type(target)}"
                )

            return OperationConsumerDecorator(self.consumer)(target)
        else:
            raise CompositionError(f"Target of unsupported type {type(target)}")



from typing import Sequence

from ..consumers import BaseURLConsumer
from .api import ServiceConsumerDecorator, ServiceDecorator

__all__: Sequence[str] = ("base_url",)


def base_url(base_url: str, /) -> ServiceDecorator:
    return ServiceConsumerDecorator(BaseURLConsumer(base_url))



from typing import Sequence

from ..consumers import HeaderConsumer
from ..enums import HTTPHeader
from .api import CommonDecorator, ConsumerDecorator

__all__: Sequence[str] = ("accept", "referer", "user_agent")


def accept(*content_types: str) -> CommonDecorator:
    return ConsumerDecorator(
        HeaderConsumer(
            HTTPHeader.ACCEPT,
            ",".join(content_types),
        )
    )


def referer(referer: str, /) -> CommonDecorator:
    return ConsumerDecorator(
        HeaderConsumer(
            HTTPHeader.REFERER,
            referer,
        )
    )


def user_agent(user_agent: str, /) -> CommonDecorator:
    return ConsumerDecorator(
        HeaderConsumer(
            HTTPHeader.USER_AGENT,
            user_agent,
        )
    )



from abc import ABC, abstractmethod
from base64 import b64encode
from dataclasses import dataclass
from typing import Sequence, TypeVar

import httpx

from .enums import HTTPHeader

__all__: Sequence[str] = (
    "Auth",
    "BasicAuth",
)

R = TypeVar("R", bound=httpx.Request)


class Auth(ABC):
    @abstractmethod
    def auth(self, request: R, /) -> R:
        raise NotImplementedError


@dataclass
class BasicAuth(Auth):
    username: str
    password: str

    def auth(self, request: R, /) -> R:
        request.headers[HTTPHeader.AUTHORIZATION] = self.authorization

        return request

    @property
    def credentials(self) -> str:
        return f"{self.username}:{self.password}"

    @property
    def token(self) -> str:
        return b64encode(self.credentials.encode()).decode()

    @property
    def authorization(self) -> str:
        return f"Basic {self.token}"



from dataclasses import dataclass
from typing import Sequence

from ..consumers import (
    CookieConsumer,
    CookiesConsumer,
    FollowRedirectsConsumer,
    HeaderConsumer,
    HeadersConsumer,
    QueryConsumer,
    QueryParamsConsumer,
    TimeoutConsumer,
    VerifyConsumer,
)
from ..operation import Operation
from ..services import ClientSpecification
from ..specification import ClientSpecification
from ..types import (
    CookiesTypes,
    CookieTypes,
    HeadersTypes,
    HeaderTypes,
    QueryParamsTypes,
    QueryTypes,
    TimeoutTypes,
    VerifyTypes,
)
from ..typing import Dependency
from .api import (
    CommonDecorator,
    ConsumerDecorator,
    SupportsConsumeClientSpecification,
    SupportsConsumeOperation,
)

__all__: Sequence[str] = (
    "cookie",
    "cookies",
    "request_depends",
    "response_depends",
    "header",
    "headers",
    "query",
    "query_params",
    "timeout",
    "verify",
)


@dataclass
class RequestDependsConsumer(
    SupportsConsumeClientSpecification, SupportsConsumeOperation
):
    dependencies: Sequence[Dependency]

    def consume_client_spec(self, client_specification: ClientSpecification) -> None:
        client_specification.request_dependencies.extend(self.dependencies)

    def consume_operation(self, operation: Operation) -> None:
        operation.request_dependencies.extend(self.dependencies)


@dataclass
class ResponseDependsConsumer(
    SupportsConsumeClientSpecification, SupportsConsumeOperation
):
    dependencies: Sequence[Dependency]

    def consume_client_spec(self, client_specification: ClientSpecification) -> None:
        client_specification.response_dependencies.extend(self.dependencies)

    def consume_operation(self, operation: Operation) -> None:
        operation.response_dependencies.extend(self.dependencies)


def request_depends(*dependencies: Dependency) -> CommonDecorator:
    return ConsumerDecorator(RequestDependsConsumer(dependencies))


def response_depends(*dependencies: Dependency) -> CommonDecorator:
    return ConsumerDecorator(ResponseDependsConsumer(dependencies))


def cookie(key: str, value: CookieTypes) -> CommonDecorator:
    return ConsumerDecorator(CookieConsumer(key, value))


def cookies(cookies: CookiesTypes, /) -> CommonDecorator:
    return ConsumerDecorator(CookiesConsumer(cookies))


def header(key: str, value: HeaderTypes) -> CommonDecorator:
    return ConsumerDecorator(HeaderConsumer(key, value))


def headers(headers: HeadersTypes, /) -> CommonDecorator:
    return ConsumerDecorator(HeadersConsumer(headers))


def query(key: str, value: QueryTypes) -> CommonDecorator:
    return ConsumerDecorator(QueryConsumer(key, value))


def query_params(params: QueryParamsTypes, /) -> CommonDecorator:
    return ConsumerDecorator(QueryParamsConsumer(params))


def timeout(timeout: TimeoutTypes, /) -> CommonDecorator:
    return ConsumerDecorator(TimeoutConsumer(timeout))


def verify(verify: VerifyTypes, /) -> CommonDecorator:
    return ConsumerDecorator(VerifyConsumer(verify))


def follow_redirects(follow_redirects: bool, /) -> CommonDecorator:
    return ConsumerDecorator(FollowRedirectsConsumer(follow_redirects))



from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import (
    Any,
    Dict,
    Generic,
    Mapping,
    MutableMapping,
    Optional,
    Sequence,
    Set,
    TypeVar,
    Union,
)

import fastapi.encoders
import httpx
from httpx import Cookies, Headers, QueryParams
from pydantic import Required
from pydantic.fields import FieldInfo, ModelField, Undefined

from .consumers import (
    CookieConsumer,
    CookiesConsumer,
    HeaderConsumer,
    HeadersConsumer,
    PathConsumer,
    PathParamsConsumer,
    QueryConsumer,
    QueryParamsConsumer,
    StateConsumer,
)
from .converters import (
    convert_cookie,
    convert_header,
    convert_path_param,
    convert_path_params,
    convert_query_param,
)
from .errors import CompositionError, ResolutionError
from .models import PreRequest, Response, State
from .resolvers import (
    BodyResolver,
    CookieResolver,
    CookiesResolver,
    HeaderResolver,
    HeadersResolver,
    QueryParamsResolver,
    QueryResolver,
    StateResolver,
)
from .types import CookiesTypes, HeadersTypes, PathParamsTypes, QueryParamsTypes
from .typing import RequestConsumer, RequestResolver, ResponseResolver, Supplier
from .utils import parse_obj_as

__all__: Sequence[str] = (
    "QueryParameter",
    "HeaderParameter",
    "CookieParameter",
    "PathParameter",
    "QueryParamsParameter",
    "HeadersParameter",
    "CookiesParameter",
    "PathParamsParameter",
    "BodyParameter",
    "URLParameter",
    "ResponseParameter",
    "RequestParameter",
    "StatusCodeParameter",
    "ReasonParameter",
    "AllRequestStateParameter",
    "AllResponseStateParameter",
    "AllStateParameter",
)

K = TypeVar("K")
V = TypeVar("V")


@dataclass(unsafe_hash=True)
class Parameter(FieldInfo):
    alias: Optional[str] = None
    default: Any = Undefined
    default_factory: Optional[Supplier[Any]] = None
    title: Optional[str] = field(default=None, compare=False)
    description: Optional[str] = field(default=None, compare=False)
    exclude: Union[Set[Union[int, str]], Mapping[Union[int, str], Any], Any] = field(
        default=None, compare=False
    )
    include: Union[Set[Union[int, str]], Mapping[Union[int, str], Any], Any] = field(
        default=None, compare=False
    )
    const: Optional[bool] = field(default=None, compare=False)
    gt: Optional[float] = field(default=None, compare=False)
    ge: Optional[float] = field(default=None, compare=False)
    lt: Optional[float] = field(default=None, compare=False)
    le: Optional[float] = field(default=None, compare=False)
    multiple_of: Optional[float] = field(default=None, compare=False)
    allow_inf_nan: Optional[bool] = field(default=None, compare=False)
    max_digits: Optional[int] = field(default=None, compare=False)
    decimal_places: Optional[int] = field(default=None, compare=False)
    min_items: Optional[int] = field(default=None, compare=False)
    max_items: Optional[int] = field(default=None, compare=False)
    unique_items: Optional[bool] = field(default=None, compare=False)
    min_length: Optional[int] = field(default=None, compare=False)
    max_length: Optional[int] = field(default=None, compare=False)
    allow_mutation: bool = field(default=True, compare=False)
    regex: Optional[str] = field(default=None, compare=False)
    discriminator: Optional[str] = field(default=None, compare=False)
    repr: bool = field(default=True, compare=False)
    extra: Dict[str, Any] = field(default_factory=dict, compare=False)
    alias_priority: Optional[int] = field(
        init=False, repr=False, default=None, compare=False
    )

    def __post_init__(self) -> None:
        # Pydantic validates that the alias is a "strict" string. This means
        # that enums (e.g. `HeaderName`) are unable to be used.
        # To mitigate this, the alias is explicity converted to a string here.
        if self.alias is not None:
            self.alias = str(self.alias)

    def compose(self, request: PreRequest, argument: Any, /) -> None:
        raise CompositionError(f"Parameter {type(self)!r} is not composable")

    def resolve_response(self, response: Response, /) -> Any:
        raise ResolutionError(
            f"Parameter {type(self)!r} is not resolvable for type {Response!r}"
        )

    def resolve_request(self, request: PreRequest, /) -> Any:
        raise ResolutionError(
            f"Parameter {type(self)!r} is not resolvable for type {PreRequest!r}"
        )

    def prepare(self, model_field: ModelField, /) -> None:
        if self.alias is None:
            self.alias = model_field.name


class ComposableSingletonParameter(ABC, Parameter, Generic[K, V]):
    def compose(self, request: PreRequest, argument: Any, /) -> None:
        if self.alias is None:
            raise CompositionError(
                f"Cannot compose parameter {type(self)!r} without an alias"
            )

        # If no argument was provided, and a value is not required for this
        # parameter, skip composition.
        if argument is None and self.default is not Required:
            return

        consumer: RequestConsumer = self.build_consumer(
            self.parse_key(self.alias),
            self.parse_value(argument),
        )

        consumer(request)

    @abstractmethod
    def parse_key(self, key: str, /) -> K:
        ...

    @abstractmethod
    def parse_value(self, value: Any, /) -> V:
        ...

    @abstractmethod
    def build_consumer(self, key: K, value: V) -> RequestConsumer:
        ...


class ResolvableSingletonParameter(ABC, Parameter, Generic[K, V]):
    def resolve_request(self, request: PreRequest, /) -> V:
        if self.alias is None:
            raise ResolutionError(
                f"Cannot resolve parameter {type(self)!r} without an alias"
            )

        resolver: RequestResolver[V] = self.build_request_resolver(
            self.parse_key(self.alias),
        )

        return resolver(request)

    def resolve_response(self, response: Response, /) -> V:
        if self.alias is None:
            raise ResolutionError(
                f"Cannot resolve parameter {type(self)!r} without an alias"
            )

        resolver: ResponseResolver[V] = self.build_response_resolver(
            self.parse_key(self.alias),
        )

        return resolver(response)

    @abstractmethod
    def parse_key(self, key: str, /) -> K:
        ...

    @abstractmethod
    def build_request_resolver(self, key: K) -> RequestResolver[V]:
        ...

    @abstractmethod
    def build_response_resolver(self, key: K) -> ResponseResolver[V]:
        ...


class ComposableSingletonStringParameter(ComposableSingletonParameter[str, str]):
    def parse_key(self, key: str, /) -> str:
        return key


class ResolvableSingletonStringParameter(
    ResolvableSingletonParameter[str, Optional[str]]
):
    def parse_key(self, key: str, /) -> str:
        return key


class QueryParameter(
    ComposableSingletonParameter[str, Sequence[str]],
    ResolvableSingletonParameter[str, Optional[Sequence[str]]],
):
    def parse_key(self, key: str, /) -> str:
        return key

    def parse_value(self, value: Any, /) -> Sequence[str]:
        return convert_query_param(value)

    def build_consumer(self, key: str, value: Sequence[str]) -> RequestConsumer:
        return QueryConsumer(key, value).consume_request

    def build_request_resolver(
        self, key: str
    ) -> RequestResolver[Optional[Sequence[str]]]:
        return QueryResolver(key).resolve_request

    def build_response_resolver(
        self, key: str
    ) -> ResponseResolver[Optional[Sequence[str]]]:
        return QueryResolver(key).resolve_response


@dataclass(unsafe_hash=True)
class HeaderParameter(
    ComposableSingletonParameter[str, Sequence[str]],
    ResolvableSingletonParameter[str, Optional[Sequence[str]]],
):
    convert_underscores: bool = True

    def parse_key(self, key: str, /) -> str:
        if self.convert_underscores:
            return key.replace("_", "-")

        return key

    def parse_value(self, value: Any, /) -> Sequence[str]:
        return convert_header(value)

    def build_consumer(self, key: str, value: Sequence[str]) -> RequestConsumer:
        return HeaderConsumer(key, value).consume_request

    def build_request_resolver(
        self, key: str
    ) -> RequestResolver[Optional[Sequence[str]]]:
        return HeaderResolver(key).resolve_request

    def build_response_resolver(
        self, key: str
    ) -> ResponseResolver[Optional[Sequence[str]]]:
        return HeaderResolver(key).resolve_response


class CookieParameter(
    ComposableSingletonStringParameter, ResolvableSingletonStringParameter
):
    def parse_value(self, value: Any, /) -> str:
        return convert_cookie(value)

    def build_consumer(self, key: str, value: str) -> RequestConsumer:
        return CookieConsumer(key, value).consume_request

    def build_request_resolver(self, key: str) -> RequestResolver[Optional[str]]:
        return CookieResolver(key).resolve_request

    def build_response_resolver(self, key: str) -> ResponseResolver[Optional[str]]:
        return CookieResolver(key).resolve_response


@dataclass(unsafe_hash=True)
class PathParameter(ComposableSingletonStringParameter):
    delimiter: str = "/"

    def parse_value(self, value: Any, /) -> str:
        return convert_path_param(value, delimiter=self.delimiter)

    def build_consumer(self, key: str, value: str) -> RequestConsumer:
        return PathConsumer(key, value).consume_request


class QueryParamsParameter(Parameter):
    def compose(self, request: PreRequest, argument: Any, /) -> None:
        params: QueryParamsTypes = parse_obj_as(QueryParamsTypes, argument)  # type: ignore

        QueryParamsConsumer(params).consume_request(request)

    def resolve_request(self, request: PreRequest, /) -> QueryParams:
        return QueryParamsResolver().resolve_request(request)

    def resolve_response(self, response: Response, /) -> QueryParams:
        return QueryParamsResolver().resolve_response(response)


class HeadersParameter(Parameter):
    def compose(self, request: PreRequest, argument: Any, /) -> None:
        headers: HeadersTypes = parse_obj_as(HeadersTypes, argument)  # type: ignore

        HeadersConsumer(headers).consume_request(request)

    def resolve_request(self, request: PreRequest, /) -> Headers:
        return HeadersResolver().resolve_request(request)

    def resolve_response(self, response: Response, /) -> Headers:
        return HeadersResolver().resolve_response(response)


class CookiesParameter(Parameter):
    def compose(self, request: PreRequest, argument: Any, /) -> None:
        cookies: CookiesTypes = parse_obj_as(CookiesTypes, argument)  # type: ignore

        CookiesConsumer(cookies).consume_request(request)

    def resolve_request(self, request: PreRequest, /) -> Cookies:
        return CookiesResolver().resolve_request(request)

    def resolve_response(self, response: Response, /) -> Cookies:
        return CookiesResolver().resolve_response(response)


@dataclass(unsafe_hash=True)
class PathParamsParameter(Parameter):
    delimiter: str = "/"

    def compose(self, request: PreRequest, argument: Any, /) -> None:
        raw_path_params: PathParamsTypes = parse_obj_as(PathParamsTypes, argument)  # type: ignore

        path_params: Mapping[str, str] = convert_path_params(
            raw_path_params, delimiter=self.delimiter
        )

        PathParamsConsumer(path_params).consume_request(request)

    def resolve_request(self, request: PreRequest) -> MutableMapping[str, str]:
        return request.path_params


@dataclass(unsafe_hash=True)
class BodyParameter(Parameter):
    embed: bool = False

    def compose(self, request: PreRequest, argument: Any, /) -> None:
        # If the parameter is not required and has no value, it can be omitted
        if argument is None and self.default is not Required:
            return

        json_value: Any = fastapi.encoders.jsonable_encoder(argument)

        if self.embed:
            if self.alias is None:
                raise CompositionError(
                    f"Cannot embed parameter {type(self)!r} without an alias"
                )

            json_value = {self.alias: json_value}

        # If this parameter shouln't be embedded in any pre-existing json,
        # make it the entire JSON request body
        if not self.embed:
            request.json = json_value
        else:
            if request.json is None:
                request.json = json_value
            else:
                request.json.update(json_value)

    def resolve_response(self, response: Response, /) -> Any:
        return BodyResolver()(response)


class URLParameter(Parameter):
    def resolve_request(self, request: PreRequest, /) -> httpx.URL:
        return request.url

    def resolve_response(self, response: Response, /) -> httpx.URL:
        return response.request.url


class ResponseParameter(Parameter):
    def resolve_response(self, response: Response, /) -> Response:
        return response


class RequestParameter(Parameter):
    def resolve_response(self, response: Response, /) -> httpx.Request:
        return response.request

    def resolve_request(self, request: PreRequest, /) -> PreRequest:
        return request


class StatusCodeParameter(Parameter):
    def resolve_response(self, response: Response, /) -> int:
        return response.status_code


class StateParameter(
    ComposableSingletonParameter[str, Any], ResolvableSingletonParameter[str, Any]
):
    def parse_key(self, key: str, /) -> str:
        return key

    def parse_value(self, value: Any, /) -> Any:
        return value

    def build_consumer(self, key: str, value: Any) -> RequestConsumer:
        return StateConsumer(key, value).consume_request

    def build_request_resolver(self, key: str) -> RequestResolver[Any]:
        return StateResolver(key).resolve_request

    def build_response_resolver(self, key: str) -> ResponseResolver[Any]:
        return StateResolver(key).resolve_response


class ReasonParameter(Parameter):
    def resolve_response(self, response: Response, /) -> str:
        return response.reason_phrase


class AllRequestStateParameter(Parameter):
    def resolve_request(self, request: PreRequest, /) -> State:
        return request.state

    def resolve_response(self, response: Response, /) -> State:
        return response.request.state


class AllResponseStateParameter(Parameter):
    def resolve_response(self, response: Response, /) -> State:
        return response.state


class AllStateParameter(AllResponseStateParameter, AllRequestStateParameter):
    pass
